/*
 *
------------------------------------------------------------
	David Olofson's entry for Comprez2000
	v0.0.1
------------------------------------------------------------
 *	Copyright (C) David Olofson, 2000
 *	This code is released under the terms of the GPL
 */

/*
TODO:
	* Real pitch modulation for the oscillators. The current event
	  controlled model is messy and inefficient, as it splits the
	  buffers for the entire part; all voices!

	* Polyphonic and Legato modes.

	* Resonant filters! 8-)

	* Better reverb for the Reverb bus.

	* Multi-tap stereo delay for the Delay bus.

	* Multi voice pitch modulator for the Chorus bus.
	  Use linear interpolation, and perhaps apply a fixed high
	  degree FIR filter before the interpolation (only one filter
	  per channel, of course!), to eliminate artifacts. (Perhaps
	  a simple 6 dB IIR filter somewhere around the 10-20 kHz
	  octave will do? Chorus FX, backing vocals etc are usually
	  cut in the treble anyway...)
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


/*====================================================================
 --------------------------------------------------------------------
	THE ENGINE
 --------------------------------------------------------------------

	This is the "audio toolkit" used to construct the
	sounds and effects, and to control them in a
	manageable fashion. It's quite similar to the core
	of a plugin based modular software synthesizer,
	although slightly more geared towards ease of use
	than towards real time audio processing.

====================================================================*/

/* Engine configuration */
#define	SAMPLE_RATE		44100
#define	BUFFER_FRAMES		1024
#define	WAVETABLE_FRAMES	2048
#define OSC_DO_INTERPOLATION
#define OSC_DO_CROSSFADE

/* Master out levels */
#define	MASTER_LEFT		0.5
#define	MASTER_RIGHT		0.5

/* Master effect mixdown levels */
#define	MASTER_REVERB		0.5
#define	MASTER_CHORUS		0.5
#define	MASTER_DELAY		0.5

/* Reverb settings */
#define	REVERB_CUTOFF		0.6
#define	REVERB_FEEDBACK		0.2
#define	REVERB_FB_CUTOFF	0.5

#define	OCTAVES			11
#define	MAX_VOICE_BUFFERS	8

/* Math constants */
#define PI			3.1415926535929323
#define	TWELWTH_ROOT_2		1.0594630944
#define	LOW_C_FREQ		(32.703196 / 4)


/*--------------------------------------------------------------------
	Globals
--------------------------------------------------------------------*/
/* Bus assignments */
#define	BUS_DRY_L	0
#define	BUS_DRY_R	1
#define	BUS_REVERB_L	2
#define	BUS_REVERB_R	3
#define	BUS_CHORUS_L	4
#define	BUS_CHORUS_R	5
#define	BUS_DELAY_L	6
#define	BUS_DELAY_R	7
#define	BUSSES		8

float	*bus[BUSSES];
float	*zero_buffer;
float	*crap_buffer;

int	engine_play_position = 0;


/*--------------------------------------------------------------------
	Pitch and frequency stuff
--------------------------------------------------------------------*/
/* semitone pitch -> oscillator k (ramp slope) table */
int	osc_k_table[128];

int get_osc_k(int pitch)
{
	int p;
	int bend = pitch & 0xff;
	if(pitch < 0)
		pitch = 0;
	pitch >>= 8;
	if(pitch > 12*OCTAVES-1)
		pitch = 12*OCTAVES-1;
	p = osc_k_table[pitch & 0x7f] * (255-bend);
	++pitch;
	p += osc_k_table[pitch & 0x7f] * bend;
	return p >> 8;
}

int get_osc_octave(int pitch)
{
	return pitch / (12*256);
}


/*--------------------------------------------------------------------
	Buffer management
--------------------------------------------------------------------*/
float *new_buffer(void)
{
	float *b;
	b = calloc(BUFFER_FRAMES, sizeof(float));
	if(!b)
	{
		fprintf(stderr, "Failed to allocate buffer!\n");
		exit(1);
	}
	return b;
}

void delete_buffer(float *buffer)
{
	free(buffer);
}


/*--------------------------------------------------------------------
	Event
--------------------------------------------------------------------*/
typedef enum
{
	/* End of track */
	evEnd = 0,
	/* Note control */
	evNoteOn,
	evNoteOff,
	/* Voice control */
	evBend,
	evMod,
	evCutOff,
	/* Voice volume */
	evVolume,
	/* Part mixing levels */
	evDry,
	evReverb,
	evChorus,
	evDelay
} events_t;

/* Event */
typedef struct event_t
{
	struct event_t	*next;
	events_t	kind;
	int		timestamp;
	char		pitch;
	float		value;
} event_t;


/*--------------------------------------------------------------------
	Wavetable
--------------------------------------------------------------------*/
typedef struct wavetable_t
{
	struct wavetable_t	*next;
	float			data[WAVETABLE_FRAMES];
} wavetable_t;

wavetable_t	*wavetables = 0;	/* global list */

wavetable_t *new_wavetable()
{
	wavetable_t *wt = calloc(sizeof(wavetable_t), 1);
	if(!wt)
	{
		fprintf(stderr,"Failed to allocate wavetable buffer!\n");
		exit(1);
	}
	wt->next = wavetables;
	wavetables = wt;
	return wt;
}

#if 0
/* Very cool effect! :-) */
wavetable_t *upsample_wavetable(wavetable_t *wtin)
{
	int i;
	float s;
	wavetable_t *wtout = new_wavetable();
	for(i = 0; i<WAVETABLE_FRAMES/2; ++i)
	{
		s = wtin->data[i*2];
		s += wtin->data[i*2 + 1];
		s *= 0.5;
		wtout->data[i] = s;
	}
	return wtout;
}
#endif

/* Create a FIR filter half-window */
float generate_half_filter(float *filter, int size)
{
	int i;
	float gain = 0;
	for(i = 0; i < size; ++i)
	{
		filter[i] = 1.0 + cos(PI*i/size);
		gain += filter[i];
	}
	return gain;
}

/* Generate an LP filtered copy of a wavetable */
wavetable_t *lpfilter_wavetable(wavetable_t *wtin, int winsize)
{
	int i, j;
	float s;
	wavetable_t *wtout = new_wavetable();
	float gain;
	float *filter;
	if(winsize < 2)
		winsize = 2;
	filter = calloc(winsize/2, sizeof(float));
	if(!filter)
	{
		fprintf(stderr,"Failed to allocate memory for filter!\n");
		exit(1);
	}
	winsize /= 2;
	gain = generate_half_filter(filter, winsize);
	gain *= 2;		/* both halves */
	gain -= filter[0];	/* For the "lost" center coefficient */
	gain = 1.0 / gain;
	for(i = 0; i < WAVETABLE_FRAMES; ++i)
	{
		s = 0;
		for(j = 0; j < winsize; ++j)
			s += wtin->data[(i+j) & (WAVETABLE_FRAMES-1)] * filter[j];
		for(j = 1; j < winsize; ++j)
			s += wtin->data[(i-j) & (WAVETABLE_FRAMES-1)] * filter[j];
		wtout->data[i] = s * gain;
	}
	free(filter);
	return wtout;
}

void cleanup_wavetables()
{
	while(wavetables)
	{
		wavetable_t *wt = wavetables;
		wavetables = wt->next;
		free(wt);
	}
}


/*--------------------------------------------------------------------
	Waveset
--------------------------------------------------------------------*/
typedef struct waveset_t
{
	wavetable_t	*table[OCTAVES];
} waveset_t;


/*--------------------------------------------------------------------
	Envelope
--------------------------------------------------------------------*/
typedef struct env_section_t
{
	struct env_section_t	*next;
	int			duration;
	float			level;
} env_section_t;

typedef struct envelope_t
{
	struct envelope_t	*next;
	env_section_t		*start;
	env_section_t		*stop;
	env_section_t		*sustain_start;
	env_section_t		*sustain_end;
	env_section_t		*release;
} envelope_t;

envelope_t *envelopes = 0;	/* global list */

envelope_t *new_envelope()
{
	envelope_t *e = calloc(sizeof(envelope_t), 1);
	if(!e)
	{
		fprintf(stderr,"Failed to allocate envelope header!\n");
		exit(1);
	}
	e->next = envelopes;
	envelopes = e;
	return e;
}

env_section_t *add_envelope_section(envelope_t *envelope)
{
	env_section_t *s = envelope->start;
	env_section_t *ns;
	ns = calloc(sizeof(env_section_t),1);
	if(!ns)
	{
		fprintf(stderr,"Failed to allocate envelope section!\n");
		exit(1);
	}
	if(s)
	{
		while(s->next)
			s = s->next;
		s->next = ns;
	}
	else
		envelope->start = ns;
	return ns;
}

void __delete_section(env_section_t *section)
{
	while(section->next)
		__delete_section(section->next);
}

void cleanup_envelope(envelope_t *envelope)
{
	__delete_section(envelope->start);
	envelope->start = 0;
}

void cleanup_envelopes()
{
	while(envelopes)
	{
		envelope_t *e = envelopes;
		envelopes = e->next;
		free(e);
	}
}



/*--------------------------------------------------------------------
	Element
--------------------------------------------------------------------*/
/* Oscillator */
typedef struct
{
	/* parameters */
	waveset_t	*waveset;
	int		syncstart;	/* evNoteOn ==> reset; may click! */
	int		pitch;		/* Controlled by evNoteOn */
	int		bend;		/* Controlled by evBend */
	int		offset;
	/* variables */
	int		real_pitch;	/* pitch + bend + offset */
	int		pos;
	int		k;
} osc_data_t;

/* Filters */
typedef struct
{
	/* parameters */
	float	r[2];	/* "resistors" */
	/* variables */
	float	c[4];	/* "capacitors" */
} filter_data_t;

/* Envelope */
typedef struct
{
	/* parameters */
	envelope_t	*envelope;
	float		scale;
	float		offset;
	/* variables */
	env_section_t	*section;
	int		timer;
	float		level;
	float		dlevel;
} envelope_data_t;

/* Mixer */
typedef struct
{
	/* parameters */
	float	old;
	float	mix;
} mixer_data_t;

/* Send */
typedef struct
{
	/* parameters */
	float	left;
	float	right;
} send_data_t;

/* Velomul */
typedef struct
{
	/* parameters */
	float	level;
} velomul_data_t;

/* Clip */
typedef struct
{
	/* parameters */
	float	gain;
	float	min;
	float	max;
} clip_data_t;

/* Union of all the above crap */
typedef union
{
	osc_data_t	osc;
	filter_data_t	filter;
	envelope_data_t	envelope;
	mixer_data_t	mixer;
	clip_data_t	clip;
	send_data_t	send;
	velomul_data_t	velomul;
} element_data_t;

struct element_t;
/*
 * part_process() can break buffers for event handling at any time for event
 * processing. As we'd prefer not to deal with that above the part level
 * where the splitting is done, and this w/o copying, the xoffset is needed
 * by the send elements so that they can calculate start positions according
 * to current position, rather than internal frame index.
 *
 * NOTE: The xoffset argument is *ONLY* for elements that access buffers
 *       external to the voice instance! The voice buffers use only the
 *       <frames> first samples, where <frames> is usually the number of
 *       sample frames until the next event for this part.
 */
typedef void (*e_process_t)(struct element_t* element, int xoffset, int frames);
typedef void (*e_process_event_t)(struct element_t* element, event_t *event);

typedef struct element_t
{
	struct element_t	*next;
	e_process_t		process;
	e_process_event_t	process_event;
	element_data_t		data;
	float			*left_in;
	float			*right_in;
	float			*control_in;
	float			*left_out;
	float			*right_out;
	float			*control_out;
} element_t;

void e_set_handlers(element_t *e, e_process_t proc, e_process_event_t proc_ev)
{
	e->process = proc;
	e->process_event = proc_ev;
}

void e_set_inputs(element_t *e, float *leftmono, float *right, float *control)
{
	if(leftmono)
		e->left_in = leftmono;
	else
		e->left_in = zero_buffer;
	if(right)
		e->right_in = right;
	else
		e->right_in = zero_buffer;
	if(control)
		e->control_in = control;
	else
		e->control_in = zero_buffer;
}

void e_set_outputs(element_t *e, float *leftmono, float *right, float *control)
{
	if(leftmono)
		e->left_out = leftmono;
	else
		e->left_out = crap_buffer;
	if(right)
		e->right_out = right;
	else
		e->right_out = crap_buffer;
	if(control)
		e->control_out = control;
	else
		e->control_out = crap_buffer;
}

/* Generic element constructor */
element_t *new_element()
{
	element_t *e;
	e = calloc(sizeof(element_t),1);
	if(!e)
	{
		fprintf(stderr, "Failed to allocate memory in new_element()\n");
	}
	e_set_handlers(e, 0, 0);
	e_set_inputs(e, 0, 0, 0);
	e_set_outputs(e, 0, 0, 0);
	return e;
}


/*--------------------------------------------------------------------
	Voice
--------------------------------------------------------------------*/
struct voice_t;

typedef void (*v_process_event_t)(struct voice_t* voice, event_t *event);

typedef struct voice_t
{
	struct voice_t		*next;
	int			playing;	/* Steal or remove if not */
	v_process_event_t	process_event;
	element_t		*elements;
	float			*buffer[MAX_VOICE_BUFFERS];
} voice_t;

void default_v_process_event(struct voice_t* voice, event_t *event)
{
	element_t *e = voice->elements;
	while(e)
	{
		if(e->process_event)
			e->process_event(e, event);
		e = e->next;
	}
}

voice_t *new_voice()
{
	voice_t *v;
	v = calloc(sizeof(voice_t),1);
	if(!v)
	{
		fprintf(stderr, "Failed to allocate memory in new_voice()\n");
		exit(1);
	}
	v->playing = 1;
	v->process_event = default_v_process_event;
	v->elements = 0;
	return v;
}

void cleanup_voice(voice_t *voice)
{
	int i;
	for(i = 0; i<MAX_VOICE_BUFFERS; ++i)
		if(voice->buffer[i])
			delete_buffer(voice->buffer[i]);
}

float *voice_buffer(voice_t *voice, int buf)
{
	if(!voice->buffer[buf])
		voice->buffer[buf] = new_buffer();
	return voice->buffer[buf];
}

void voice_clear_buffers(voice_t *voice, int frames)
{
	int i;
	for(i = 0; i<MAX_VOICE_BUFFERS; ++i)
		if(voice->buffer[i])
			memset(voice->buffer[i], 0, frames * sizeof(float));
}

void voice_process(voice_t *voice, int xoffset, int frames)
{
	element_t *e = voice->elements;
	voice_clear_buffers(voice, frames);
	while(e)
	{
		if(e->process)
			e->process(e, xoffset, frames);
		e = e->next;
	}
}


/*====================================================================
	Voice construction tools
====================================================================*/
element_t *add_element(voice_t *voice)
{
	element_t *ne;
	element_t *e = voice->elements;
	ne = new_element();
	if(!e)
		voice->elements = ne;
	else
	{
		while(e->next)
			e = e->next;
		e->next = ne;
	}
	return ne;
}

/*---------------------------------
	Oscillator
---------------------------------*/
void osc_process(struct element_t* element, int xoffset, int frames)
{
	osc_data_t *data = &element->data.osc;
	float *wave, *wave2;
	float s, f;
	int p;
	float *out = element->left_out;
	int oct = get_osc_octave(data->real_pitch);
#ifdef OSC_DO_CROSSFADE
	float s2, octf;
#endif
	wave = data->waveset->table[oct]->data;
#ifdef OSC_DO_CROSSFADE
	++oct;
	if(oct > OCTAVES - 1)
		oct = OCTAVES - 1;
	wave2 = data->waveset->table[oct]->data;
	octf = (data->real_pitch) % (12<<8);
	octf /= 12*256;
#endif
	while(frames--)
	{
		p = data->pos>>8;
#ifdef OSC_DO_INTERPOLATION
		f = (data->pos & 0xff) * 0.00390625;
#ifdef OSC_DO_CROSSFADE
		s = wave[p & (WAVETABLE_FRAMES - 1)] * (1.0-f);
		s2 = wave2[p & (WAVETABLE_FRAMES - 1)] * (1.0-f);
		++p;
		s += wave[p & (WAVETABLE_FRAMES - 1)] * f;
		s2 += wave2[p & (WAVETABLE_FRAMES - 1)] * f;
		s *= 1.0 - octf;
		s2 *= octf;
		s += s2;
#else
		s = wave[p & (WAVETABLE_FRAMES - 1)] * (1.0-f);
		++p;
		s += wave[p & (WAVETABLE_FRAMES - 1)] * f;
#endif
#else
#ifdef OSC_DO_CROSSFADE
		s = wave[p & (WAVETABLE_FRAMES - 1)] * (1.0 - octf);
		s += wave2[p & (WAVETABLE_FRAMES - 1)] * octf;
#else
		s = wave[p & (WAVETABLE_FRAMES - 1)];
#endif
#endif
		data->pos += data->k;
		*out++ = s;
	}	
}

void osc_process_add(struct element_t* element, int xoffset, int frames)
{
	osc_data_t *data = &element->data.osc;
	float *wave, *wave2;
	float s, f;
	int p;
	float *out = element->left_out;
	int oct = get_osc_octave(data->real_pitch);
	float s2, octf;
	wave = data->waveset->table[oct]->data;
	++oct;
	if(oct > OCTAVES - 1)
		oct = OCTAVES - 1;
	wave2 = data->waveset->table[oct]->data;
	octf = (data->real_pitch) % (12<<8);
	octf /= 12*256;
	while(frames--)
	{
		p = data->pos>>8;
		f = (data->pos & 0xff) * 0.00390625;
		s = wave[p & (WAVETABLE_FRAMES - 1)] * (1.0-f);
		s2 = wave2[p & (WAVETABLE_FRAMES - 1)] * (1.0-f);
		++p;
		s += wave[p & (WAVETABLE_FRAMES - 1)] * f;
		s2 += wave2[p & (WAVETABLE_FRAMES - 1)] * f;
		s *= 1.0 - octf;
		s2 *= octf;
		s += s2;
		s *= 0.5;
		data->pos += data->k;
		*out++ += s;
	}	
}

void osc_process_event(struct element_t* element, event_t *event)
{
	osc_data_t *data = &element->data.osc;
	switch(event->kind)
	{
	  case evEnd:
		break;
	  case evNoteOn:
//fprintf(stderr, "evNoteOn, pitch = %d, ", event->pitch);
		data->pitch = event->pitch << 8;
		if(data->syncstart)
			data->pos = 0;
//fprintf(stderr, "data->k = %d\n", data->k);
		break;
	  case evNoteOff:
		break;
	  case evBend:
		data->bend = event->value * 256.0;
		break;
	  case evMod:
	  case evCutOff:
	  case evVolume:
	  case evDry:
	  case evReverb:
	  case evChorus:
	  case evDelay:
		break;
	}
	data->real_pitch = data->pitch + data->bend + data->offset;
	data->k = get_osc_k(data->real_pitch);
}

element_t *add_oscillator(voice_t *voice, waveset_t *waveset, int output,
								float offset)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, osc_process, osc_process_event);
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	e->data.osc.waveset = waveset;
	e->data.osc.offset = offset * 256.0;
	return e;
}

element_t *add_oscillator_add(voice_t *voice, waveset_t *waveset, int output,
								float offset)
{
	element_t *e = add_oscillator(voice, waveset, output, offset);
	e_set_handlers(e, osc_process_add, osc_process_event);
	return e;
}


/*---------------------------------
	Multiplier
---------------------------------*/
void mul_process(struct element_t* element, int xoffset, int frames)
{
	float *in = element->left_in;
	float *out = element->left_out;
	while(frames--)
		*out++ *= *in++;
}

element_t *add_multiplier(voice_t *voice, int input, int output)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, mul_process, 0);
	e_set_inputs(e, voice_buffer(voice, input), 0, 0);
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	return e;
}


/*---------------------------------
	Velocity Multiplier
---------------------------------*/
void velomul_process(struct element_t* element, int xoffset, int frames)
{
	float *in = element->left_in;
	float *out = element->left_out;
	float lev = element->data.velomul.level;
	while(frames--)
		*out++ *= *in++ * lev;
}

void velomul_process_event(struct element_t* element, event_t *event)
{
	velomul_data_t *data = &element->data.velomul;
	switch(event->kind)
	{
	  case evNoteOn:
		data->level = event->value;
	  default:
		break;
	}
}

element_t *add_velomul(voice_t *voice, int input, int output)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, velomul_process, velomul_process_event);
	e_set_inputs(e, voice_buffer(voice, input), 0, 0);
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	return e;
}


/*---------------------------------
	Mixer
---------------------------------*/
void mix_process(struct element_t* element, int xoffset, int frames)
{
	float *in = element->left_in;
	float *out = element->left_out;
	float oldlev = element->data.mixer.old;
	float mixlev = element->data.mixer.mix;
	while(frames--)
	{
		*out = *out*oldlev + *in++*mixlev;
		++out;
	}
}

element_t *add_mixer(voice_t *voice, int input, int output,
						float mixlev, float oldlev)
{
	element_t *e;
	mixer_data_t *data;
	e = add_element(voice);
	data = &e->data.mixer;
	e_set_handlers(e, mix_process, 0);
	e_set_inputs(e, voice_buffer(voice, input), 0, 0);
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	data->old = oldlev;
	data->mix =mixlev;
	return e;
}


/*---------------------------------
	Adder
---------------------------------*/
void adder_process(struct element_t* element, int xoffset, int frames)
{
	float *in = element->left_in;
	float *out = element->left_out;
	float mixlev = element->data.mixer.mix;
	while(frames--)
		*out++ += *in++*mixlev;
}

element_t *add_adder(voice_t *voice, int input, int output, float level)
{
	element_t *e;
	mixer_data_t *data;
	e = add_element(voice);
	data = &e->data.mixer;
	e_set_handlers(e, adder_process, 0);
	e_set_inputs(e, voice_buffer(voice, input), 0, 0);
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	data->mix = level;
	return e;
}


/*---------------------------------
	Send
---------------------------------*/
typedef enum
{
	sDry = 0,
	sReverb,
	sChorus,
	sDelay
} send_bus_t;

void send_process(struct element_t* element, int xoffset, int frames)
{
	float *in = element->left_in;
	float *lout = element->left_out + xoffset;
	float *rout = element->right_out + xoffset;
	float leftlev = element->data.send.left;
	float rightlev = element->data.send.right;
	while(frames--)
	{
		*lout++ += *in*leftlev;
		*rout++ += *in*rightlev;
		++in;
	}
}

element_t *add_send(voice_t *voice, int from, send_bus_t sbus,
						float vol, float pan)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, send_process, 0);
	e_set_inputs(e, voice_buffer(voice, from), 0, 0);
	e_set_outputs(e, bus[sbus*2], bus[sbus*2 + 1], 0);
	e->data.send.left = vol - pan;
	e->data.send.right = vol + pan;
	return e;
}


/*---------------------------------
	Envelope
---------------------------------*/

void env_next_section(envelope_data_t* env);

void env_set_section(envelope_data_t* env, env_section_t *section)
{
	if(section)
	{
		float lev = section->level * env->scale + env->offset;
		env->section = section;
		if(section->duration)
		{
			env->dlevel = (lev - env->level)
						/ section->duration;
			env->timer = section->duration;
		}
		else
		{
			env->level = lev;
			env->timer = 0;
		}
	}
	else
	{
		env->section = 0;
		env->dlevel = 0;
		env->timer = 1000000000;
	}
}

void env_start(envelope_data_t* env)
{
	/*env->level = 0;*/
	env_set_section(env, env->envelope->start);
}

void env_release(envelope_data_t* env)
{
	env_set_section(env, env->envelope->release);
}

void env_next_section(envelope_data_t* env)
{
	if(!env->section)
	{
		env->timer = 1000000000;
		return;
	}
	if(env->section == env->envelope->stop)
	{
		env->section = 0;
		env->dlevel = 0;
		env->timer = 1000000000;
		return;
	}
	env->level = env->section->level * env->scale + env->offset;
	env_set_section(env, env->section->next);
}

void env_process(struct element_t* element, int xoffset, int frames)
{
	envelope_data_t *data = &element->data.envelope;
	float *out = element->control_out;
	while(frames--)
	{
		data->level += data->dlevel;
		/* while(), since we support duration == 0 */
		while(!data->timer--)
			env_next_section(data);
		*out++ = data->level;
	}
}

void env_process_event(struct element_t* element, event_t *event)
{
	switch(event->kind)
	{
	  case evEnd:
		break;
	  case evNoteOn:
		env_start(&element->data.envelope);
		break;
	  case evNoteOff:
		env_release(&element->data.envelope);
		break;
	  case evBend:
		break;
	  case evMod:
	  case evCutOff:
	  case evVolume:
	  case evDry:
	  case evReverb:
	  case evChorus:
	  case evDelay:
		break;
	}
}

element_t *add_envelope(voice_t *voice, envelope_t *env, int output)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, env_process, env_process_event);
	e_set_outputs(e, 0, 0, voice_buffer(voice, output));
	e->data.envelope.envelope = env;
	e->data.envelope.scale = 1.0;
	return e;
}


/*---------------------------------
	Simple 6dB LP filter
---------------------------------*/
void lpf_process(struct element_t* element, int xoffset, int frames)
{
	filter_data_t *d = &element->data.filter;
	float *ctrl = element->control_in;
	float *in = element->left_in;
	float *out = element->left_out;
	while(frames--)
	{
		d->c[0] += (*in++ - d->c[0]) * *ctrl++ * d->r[0];
		*out++ = d->c[0];
	}
}

element_t *add_lpf(voice_t *voice, int input, int output, int control,
								float fcoeff)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, lpf_process, 0);
	e_set_inputs(e, voice_buffer(voice, input), 0,
					voice_buffer(voice, control));
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	e->data.filter.r[0] = fcoeff;
	return e;
}


/*---------------------------------
	Simple 12dB LP filter
---------------------------------*/
void lpf12_process(struct element_t* element, int xoffset, int frames)
{
	filter_data_t *d = &element->data.filter;
	float *ctrl = element->control_in;
	float *in = element->left_in;
	float *out = element->left_out;
	float r;
	while(frames--)
	{
		r = *ctrl++ * d->r[0];
		d->c[0] += (*in++ - d->c[0]) * r;
		d->c[1] += (d->c[0] - d->c[1]) * r;
		*out++ = d->c[1];
	}
}

element_t *add_lpf12(voice_t *voice, int input, int output, int control,
								float fcoeff)
{
	element_t *e;
	e = add_element(voice);
	e_set_handlers(e, lpf_process, 0);
	e_set_inputs(e, voice_buffer(voice, input), 0,
					voice_buffer(voice, control));
	e_set_outputs(e, voice_buffer(voice, output), 0, 0);
	e->data.filter.r[0] = fcoeff;
	return e;
}


/*--------------------------------------------------------------------
	Part
--------------------------------------------------------------------*/
typedef enum
{
	vaMono,
	vaLegato,
	vaPoly
} va_mode_t;

struct part_t;
typedef voice_t *(*voice_constructor_t)(struct part_t *part);

typedef struct part_t
{
	struct part_t		*next;

	/* Event port */
	event_t			*events;
	event_t			*events_in;

	int			finished;

	/* key states */
	char			key_down[128];
	char			key_count[128];
	char			key_counter;
	int			keys_down;
	int			current_bend;

	va_mode_t		va_mode;
	voice_t			*voices;
	voice_constructor_t	voice_create;
} part_t;

part_t	*parts = 0;	/* global list */

voice_t *part_add_voice(part_t *part)
{
	voice_t	*v = part->voices;
	part->voices = part->voice_create(part);
	part->voices->next = v;
	return part->voices;
}

void part_remove_voice(part_t *part, voice_t *voice)
{
	voice_t	*v, *prevv;
	prevv = 0;
	v = part->voices;
	/* find */
	while(v != voice)
	{
		prevv = v;
		v = v->next;
		if(!v)
			return;
	}
	/* unlink */
	if(!prevv)
		part->voices = v->next;
	else
		prevv->next = v->next;
	/* delete */
	cleanup_voice(v);
	free(v);
}

part_t *new_part(voice_constructor_t voicecon, va_mode_t mode)
{
	part_t *p;
	p = calloc(sizeof(part_t),1);
	if(!p)
	{
		fprintf(stderr, "Failed to allocate memory in new_part()\n");
		exit(1);
	}
	p->voice_create = voicecon;
	p->va_mode = mode;
	switch(mode)
	{
	  case vaMono:
	  case vaLegato:
		part_add_voice(p);
	  	break;
	  case vaPoly:
	  	break;
	}
	return p;
}

void cleanup_part(part_t *part)
{
	event_t *e = part->events;
	while(e)
	{
		part->events = e->next;
		free(e);
	}
	while(part->voices)
		part_remove_voice(part, part->voices);
}

void part_process_event(part_t* part, event_t *event)
{
	switch(event->kind)
	{
	  case evEnd:
		part->finished = 1;
		break;
	  case evNoteOn:
	  case evNoteOff:
	  case evBend:
	  case evMod:
	  case evCutOff:
	  case evVolume:
	  case evDry:
	  case evReverb:
	  case evChorus:
	  case evDelay:
		break;
	}
}

void part_process(part_t *part)
{
	voice_t	*v;
	int	do_frames = BUFFER_FRAMES;
	int	frames_left = BUFFER_FRAMES;
	int	position = engine_play_position;
	int	offset = 0;
	while(1)
	{
		if(part->events)
			do_frames = part->events->timestamp - position;
		else
			do_frames = BUFFER_FRAMES;
		if(do_frames)
		{
			int no_more_events = 0;
			if(do_frames >= frames_left)
			{
				do_frames = frames_left;
				no_more_events = 1;
			}
			v = part->voices;
			while(v)
			{
				voice_process(v, offset, do_frames);
				v = v->next;
			}
			if(no_more_events)
				return;

			frames_left -= do_frames;
			offset += do_frames;
			position += do_frames;
		}
		if(part->events)
		{
			part_process_event(part, part->events);
			v = part->voices;
			while(v)
			{
				v->process_event(v, part->events);
				v = v->next;
			}
			part->events = part->events->next;
		}
	}
}


/*--------------------------------------------------------------------
	Part event management and tools
--------------------------------------------------------------------*/
event_t *__insert_event(part_t *part, event_t *e)
{
#warning Out-of-order events not yet supported!
	fprintf(stderr, "Out-of-order events not yet supported!\n");
	exit(1);
	return e;
}

event_t *part_add_event(part_t *part, int timestamp)
{
	event_t *ne;
	ne = calloc(sizeof(event_t), 1);
	ne->timestamp = timestamp;
	if(!ne)
	{
		fprintf(stderr, "Failed to allocate event!\n");
		exit(1);
	}
	if(part->events_in)
	{
		if(part->events_in->timestamp > ne->timestamp)
			return __insert_event(part, ne);
		part->events_in->next = ne;
		part->events_in = ne;
	}
	else
		part->events_in = part->events = ne;
	return ne;
}

void End(part_t *p, int timestamp)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evEnd;
}

void NoteOn(part_t *p, int timestamp, char pitch, float velocity)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evNoteOn;
	e->pitch = pitch;
	e->value = velocity;
}

void NoteOff(part_t *p, int timestamp, char pitch)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evNoteOff;
	e->pitch = pitch;
	e->value = 0.5;
}

void NoteOffVel(part_t *p, int timestamp, char pitch, float velocity)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evNoteOff;
	e->pitch = pitch;
	e->value = velocity;
}

void Bend(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evBend;
	e->value = amount;
}

void Mod(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evMod;
	e->value = amount;
}

void CutOff(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evCutOff;
	e->value = amount;
}

void Volume(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evVolume;
	e->value = amount;
}

void Dry(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evDry;
	e->value = amount;
}

void Reverb(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evReverb;
	e->value = amount;
}

void Chorus(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evChorus;
	e->value = amount;
}

void Delay(part_t *p, int timestamp, float amount)
{
	event_t *e = part_add_event(p, timestamp);
	e->kind = evDelay;
	e->value = amount;
}


part_t *add_part(voice_constructor_t voicecon, va_mode_t mode)
{
	part_t *part = new_part(voicecon, mode);
	part->next = parts;
	parts = part;
	return part;
}

void cleanup_parts()
{
	part_t *part = parts;
	while(part)
	{
		parts = part->next;
		cleanup_part(part);
	}
}

void process_parts()
{
	part_t *p = parts;
	while(p)
	{
		/* In case you forget to use it at all... */
		if(!p->events)
			p->finished = 1;
		part_process(p);
		p = p->next;
	}
}


/*--------------------------------------------------------------------
	Top level processing
--------------------------------------------------------------------*/
#define	RVB_BUFSIZE	65536
#define	RVB_BUFMASK	(RVB_BUFSIZE-2)

struct
{
	/* input filters */
	float	cl;
	float	cr;
	/* feedback filters */
	float	fbcl;
	float	fbcr;
	/* delay buffer (stereo) */
	int	bufferpos;
	float	*buffer;
} reverb;

void reverb_init()
{
	reverb.buffer = calloc(RVB_BUFSIZE, sizeof(float));
	if(!reverb.buffer)
	{
		fprintf(stderr, "Failed to allocate buffer for the reverb!\n");
		exit(1);
	}
}

void reverb_cleanup()
{
	free(reverb.buffer);
}

void reverb_process(float **inputs, float **outputs)
{
	int s = BUFFER_FRAMES;
	int i, j;
	int bp = reverb.bufferpos;
	float *inl = inputs[0];
	float *inr = inputs[1];
	float *outl = outputs[0];
	float *outr = outputs[1];
	float accul, accur;
	while(s--)
	{
		/* Input */
		reverb.cl += (*inl++ - reverb.cl) * REVERB_CUTOFF;
		reverb.cr += (*inr++ - reverb.cr) * REVERB_CUTOFF;
		reverb.buffer[bp & RVB_BUFMASK] = reverb.cl;
		reverb.buffer[(bp & RVB_BUFMASK) + 1] = reverb.cr;

		/* Feedback LP filters */
		reverb.fbcl += (reverb.buffer[((bp-24350) & RVB_BUFMASK) + 1]
				+ reverb.buffer[(bp-45800) & RVB_BUFMASK]
				- reverb.fbcl) * REVERB_FB_CUTOFF;
		reverb.fbcr += (reverb.buffer[(bp-12340) & RVB_BUFMASK]
				+ reverb.buffer[((bp-43550) & RVB_BUFMASK) + 1]
				- reverb.fbcr) * REVERB_FB_CUTOFF;

		/* Feedback */
		reverb.buffer[bp & RVB_BUFMASK] += reverb.fbcl * REVERB_FEEDBACK;
		reverb.buffer[(bp & RVB_BUFMASK) + 1] += reverb.fbcr * REVERB_FEEDBACK;
		reverb.buffer[(bp-12167) & RVB_BUFMASK] -= reverb.fbcl * REVERB_FEEDBACK * 0.5;
		reverb.buffer[((bp-24930) & RVB_BUFMASK) + 1] -= reverb.fbcr * REVERB_FEEDBACK * 0.5;
		reverb.buffer[(bp-27732) & RVB_BUFMASK] -= reverb.fbcr * REVERB_FEEDBACK * 0.5;
		reverb.buffer[((bp-16465) & RVB_BUFMASK) + 1] -= reverb.fbcl * REVERB_FEEDBACK * 0.5;

#if 0		/* Taps */
		accul = accur = 0;
		j = 3000;
		for(i = 700; i<20000; i += j)
		{
			accul += reverb.buffer[((bp-i) & RVB_BUFMASK) + 1] * 0.1;
			j += 17;
		}
		j = 4000;
		for(i = 570; i<20000; i += j)
		{
			accur += reverb.buffer[(bp-i) & RVB_BUFMASK] * 0.1;
			j += 15;
		}
		j = 2000;
		for(i = 930; i<20000; i += j)
		{
			accul += reverb.buffer[(bp-i) & RVB_BUFMASK] * 0.1;
			j += 53;
		}
		j = 3300;
		for(i = 1140; i<20000; i += j)
		{
			accur += reverb.buffer[((bp-i) & RVB_BUFMASK) + 1] * 0.1;
			j += 63;
		}
#endif

		accul = reverb.buffer[(bp-942) & RVB_BUFMASK] * 0.5;
		accur = reverb.buffer[((bp-1556) & RVB_BUFMASK) + 1] * 0.5;
/*		accul += reverb.buffer[(bp-1920) & RVB_BUFMASK] * 0.3;
		accur += reverb.buffer[((bp-3530) & RVB_BUFMASK) + 1] * 0.3;
		accul += reverb.buffer[(bp-7344) & RVB_BUFMASK] * 0.2;
		accur += reverb.buffer[((bp-9578) & RVB_BUFMASK) + 1] * 0.2;

		accul = reverb.buffer[((bp-3360) & RVB_BUFMASK) + 1];
		accur = reverb.buffer[(bp-4800) & RVB_BUFMASK] * 0.5;
		accul += reverb.buffer[(bp-7000) & RVB_BUFMASK] * 0.25;
		accur += reverb.buffer[((bp-8550) & RVB_BUFMASK) + 1] * 0.25;
		accul += reverb.buffer[(bp-22300) & RVB_BUFMASK] * 2.0;
		accur += reverb.buffer[((bp-25550) & RVB_BUFMASK) + 1] * 2.0;
		accul += reverb.buffer[(bp-37800) & RVB_BUFMASK];
		accur += reverb.buffer[((bp-39660) & RVB_BUFMASK) + 1];
*/
		*outl++ += accul * MASTER_REVERB;
		*outr++ += accur * MASTER_REVERB;

		bp += 2;
	}
	reverb.bufferpos = bp;
}

void clear_busses()
{
	int i;
	for(i = 0; i<BUSSES; ++i)
		memset(bus[i], 0, BUFFER_FRAMES * sizeof(float));
}
 
void master_mixdown()
{
	reverb_process(&bus[BUS_REVERB_L], &bus[BUS_DRY_L]);
//	chorus_process(&bus[BUS_CHORUS_L], &bus[BUS_DRY_L]);
//	delay_process(&bus[BUS_DELAY_L], &bus[BUS_DRY_L]);
}

void f2i(float *inl, float *inr, short *outstereo)
{
	char *byteout = (char *)outstereo;
	int s = BUFFER_FRAMES;
	float l, r;
	int il, ir;
	while(s--)
	{
		l = *inl++ * MASTER_LEFT * 32768;
		r = *inr++ * MASTER_RIGHT * 32768;
		if(l < -32767)
			l = -32767;
		else if(l > 32767)
			l = 32767;
		if(r < -32767)
			r = -32767;
		else if(r > 32767)
			r = 32767;
		/* Intel byte order it is!
		 * (In case someone compiles this on a little endian CPU...)
		 */
		il = l;
		ir = r;
		*byteout++ = il;
		*byteout++ = il>>8;
		*byteout++ = ir;
		*byteout++ = ir>>8;
	}
}


/*--------------------------------------------------------------------
	Toplevel engine "run" call
--------------------------------------------------------------------*/
void engine_run(short *outbuf)
{
	clear_busses();
	process_parts();
	master_mixdown();
	f2i(bus[BUS_DRY_L], bus[BUS_DRY_R], outbuf);
	engine_play_position += BUFFER_FRAMES;
}

int engine_finished()
{
	part_t *p = parts;
	while(p)
		if(!p->finished)
			return 0;
		else
			p = p->next;
	return 1;
}


/*====================================================================
	Init functions
====================================================================*/
void init_k_table(int samplerate)
{
	double f = LOW_C_FREQ;
	int i;
	for(i=0; i<128; i++)
	{
		osc_k_table[i] = floor(f*WAVETABLE_FRAMES*256.0/samplerate + 0.5);
		f *= TWELWTH_ROOT_2;
	}
}

void init_mixing()
{
	int i;
	for(i = 0; i<BUSSES; ++i)
		bus[i] = new_buffer();
	zero_buffer = new_buffer();
	crap_buffer = new_buffer();
}

void init_master_fx()
{
	reverb_init();
}

void init_engine(int rate)
{
	init_k_table(rate);
	init_mixing();
	init_mixing();
	init_master_fx();
}


/*====================================================================
	Cleanup functions
====================================================================*/
void cleanup_mixing()
{
	int i;
	for(i = 0; i<BUSSES; ++i)
		delete_buffer(bus[i]);
	delete_buffer(zero_buffer);
	delete_buffer(crap_buffer);
}

void cleanup_master_fx()
{
	reverb_cleanup();
}

void cleanup_engine()
{
	cleanup_mixing();
	cleanup_master_fx();
}



/*====================================================================
 --------------------------------------------------------------------
	THE INSTRUMENTS
 --------------------------------------------------------------------

	This is where waveform, envelopes and other instrument
	specific data is constructed.

====================================================================*/

/*====================================================================
	Wave sets and Envelopes
====================================================================*/

/*--------------------------------------------------------------------
	Waveforms
--------------------------------------------------------------------*/
waveset_t waveset_sine;
waveset_t waveset_saw;
waveset_t waveset_pulse[6];
waveset_t waveset_softpulse;

void init_oscillators()
{
	int i, j;
	int size;
	wavetable_t *wt;

	/* Generate sine wavetables */
	wt = new_wavetable();
	for(i = 0; i < WAVETABLE_FRAMES; ++i)
		wt->data[i] = sin(i*2*PI/WAVETABLE_FRAMES);
	for(i = 0; i < OCTAVES; ++i)
		waveset_sine.table[i] = wt;

	/* Generate saw wavetables */
	wt = new_wavetable();
	for(i = 0; i < WAVETABLE_FRAMES; ++i)
		wt->data[i] = 1.0 - (float)i*2.0/WAVETABLE_FRAMES;
	size = WAVETABLE_FRAMES * 1.1;
	for(i = OCTAVES-1; i >= 0; --i)
	{
		if(size < 2)
			waveset_saw.table[i] = wt;
		else
			waveset_saw.table[i] =
				lpfilter_wavetable(wt, size);
		size /= 2;
	}

	/* Generate pulse wavetables */
	for(j = 0; j < 6; ++j)
	{
		int switchpoint = WAVETABLE_FRAMES * j / 10;	/* 10..50% */
		if(!switchpoint)
			switchpoint = WAVETABLE_FRAMES * 3 / 100;	/* 3% */
		wt = new_wavetable();
		for(i = 0; i < WAVETABLE_FRAMES; ++i)
			wt->data[i] = 1.0 - 2.0*(i > switchpoint);
		size = WAVETABLE_FRAMES * 1.1;
		for(i = OCTAVES-1; i >= 0; --i)
		{
			if(size < 2)
				waveset_pulse[j].table[i] = wt;
			else
				waveset_pulse[j].table[i] =
					lpfilter_wavetable(wt, size);
			size /= 2;
		}
	}

	/* Set up "softpulse" wavetable */
	j = 5;
	for(i = 0; i < OCTAVES; ++i)
	{
		waveset_softpulse.table[i] = waveset_pulse[5].table[j];
		++j;
		if(j>OCTAVES-1)
			j = OCTAVES-1;
	}
}

void cleanup_oscillators()
{
	cleanup_wavetables();
}


/*--------------------------------------------------------------------
	Envelopes
--------------------------------------------------------------------*/
envelope_t envelope_out[4];
envelope_t envelope_BassDrum;
envelope_t envelope_HiHat;
envelope_t envelope_Bass;
envelope_t envelope_Strings;

void init_envelopes()
{
	int	i;
	envelope_t	*e;
	env_section_t	*es;

	for(i = 0; i<4; ++i)
	{
		e = &envelope_out[i];
			es = add_envelope_section(e);
				es->duration = 10;
				es->level = 1.0;
			es = add_envelope_section(e);
				es->duration = 1000 + i*500;
				es->level = 0.7;
			es = add_envelope_section(e);
			e->release = es;
				es->duration = 1500 + i*700;
				es->level = 0.0;
	}

	e = &envelope_BassDrum;
		es = add_envelope_section(e);
			es->duration = 10;
			es->level = 0.2;
		es = add_envelope_section(e);
			es->duration = 10;
			es->level = 0.5;
		es = add_envelope_section(e);
			es->duration = 2000;
			es->level = 1.0;
		es = add_envelope_section(e);
			es->duration = 3000;
			es->level = 0.9;
		es = add_envelope_section(e);
		e->stop = es;
			es->duration = 10000;
			es->level = 0.0;
		es = add_envelope_section(e);
		e->release = es;
			es->duration = 2000;
			es->level = 0.0;

	e = &envelope_Bass;
		es = add_envelope_section(e);
			es->duration = 5;
			es->level = 1.0;
		es = add_envelope_section(e);
			es->duration = 300;
			es->level = 0.6;
		es = add_envelope_section(e);
			es->duration = 2000;
			es->level = 0.5;
		es = add_envelope_section(e);
		e->stop = es;
			es->duration = 4000;
			es->level = 0.0;
		es = add_envelope_section(e);
		e->release = es;
			es->duration = 1000;
			es->level = 0.0;

	e = &envelope_HiHat;
		es = add_envelope_section(e);
			es->duration = 5;
			es->level = 1.0;
		es = add_envelope_section(e);
			es->duration = 1000;
			es->level = 0.3;
		es = add_envelope_section(e);
		e->stop = es;
			es->duration = 10000;
			es->level = 0.0;
		es = add_envelope_section(e);
		e->release = es;
			es->duration = 1500;
			es->level = 0.0;

	e = &envelope_Strings;
		es = add_envelope_section(e);
			es->duration = 200;
			es->level = 0.5;
		es = add_envelope_section(e);
			es->duration = 10000;
			es->level = 0.8;
		es = add_envelope_section(e);
		e->stop = es;
			es->duration = 40000;
			es->level = 0.7;
		es = add_envelope_section(e);
		e->release = es;
			es->duration = 20000;
			es->level = 0.0;
}


/*--------------------------------------------------------------------
	Instruments & parts
--------------------------------------------------------------------*/

part_t *p_BassDrum = 0;
part_t *p_HiHat = 0;
part_t *p_Bass = 0;
part_t *p_LoopSynth = 0;
part_t *p_Strings[6] = { 0,0,0,0,0,0 };


voice_t *BassDrum_create(part_t *part)
{
	voice_t	*v = new_voice(part);
	element_t *e;
	e = add_oscillator(v, &waveset_sine, 0, 0.0);
		e->data.osc.syncstart = 1;
	e = add_oscillator(v, &waveset_softpulse, 1, 2.0);
		e->data.osc.syncstart = 1;
	add_envelope(v, &envelope_out[2], 2);
	add_multiplier(v, 2, 1);
	add_adder(v, 1, 0, 1.0);
	add_envelope(v, &envelope_BassDrum, 1);
	add_velomul(v, 1, 0);
	add_send(v, 0, sDry, 1.0, 0.0);
	add_send(v, 0, sReverb, 0.1, 0.0);
	return v;
}

voice_t *HiHat_create(part_t *part)
{
	voice_t	*v = new_voice(part);
/*
	add_oscillator(v, &waveset_saw, 0, 60.0);
	add_oscillator(v, &waveset_saw, 1, 71.0);
	add_multiplier(v, 1, 0);
	add_oscillator(v, &waveset_saw, 1, 87.0);
	add_multiplier(v, 1, 0);
	add_oscillator(v, &waveset_saw, 1, 77.5);
	add_multiplier(v, 1, 0);
*/
	add_oscillator(v, &waveset_saw, 0, 54.5);
	add_oscillator(v, &waveset_saw, 1, 76.5);
	add_multiplier(v, 1, 0);
	add_oscillator(v, &waveset_saw, 1, 88.0);
	add_multiplier(v, 1, 0);
	add_oscillator(v, &waveset_saw, 1, 91.3);
	add_multiplier(v, 1, 0);

	add_envelope(v, &envelope_HiHat, 1);
	add_velomul(v, 1, 0);
	add_send(v, 0, sDry, 2.0, 0.0);
	add_send(v, 0, sReverb, 1.0, 0.0);
	return v;
}

voice_t *Bass_create(part_t *part)
{
	voice_t	*v = new_voice(part);
	element_t *e;
	e = add_oscillator(v, &waveset_sine, 0, 0.0);
		e->data.osc.syncstart = 1;
	e = add_oscillator(v, &waveset_softpulse, 1, -0.1);
		e->data.osc.syncstart = 1;
	e = add_oscillator_add(v, &waveset_saw, 1, 12.1);
//		e->data.osc.syncstart = 1;
	add_adder(v, 1, 0, 0.2);
	e = add_oscillator(v, &waveset_sine, 1, 11.9);
		e->data.osc.syncstart = 1;
	add_adder(v, 1, 0, 0.5);
	e = add_envelope(v, &envelope_out[0], 1);
		e->data.envelope.scale = 0.7;
		e->data.envelope.offset = 0.1;
	add_lpf(v, 0, 0, 1, 0.1);
	add_envelope(v, &envelope_Bass, 1);
	add_velomul(v, 1, 0);
	add_send(v, 0, sDry, 0.9, 0.0);
	add_send(v, 0, sReverb, 0.2, 0.0);
	return v;
}

voice_t *LoopSynth_create(part_t *part)
{
	voice_t	*v = new_voice(part);
	add_oscillator(v, &waveset_saw, 0, -0.1);
	add_oscillator(v, &waveset_saw, 1, 0.2);
	add_oscillator(v, &waveset_softpulse, 2, 12.0);
	add_mixer(v, 1, 0, 0.3, 0.3);
	add_multiplier(v, 1, 0);
	add_adder(v, 2, 0, 0.2);
	add_envelope(v, &envelope_out[3], 3);
	add_velomul(v, 3, 0);
	add_envelope(v, &envelope_out[1], 1);
	add_lpf(v, 0, 0, 1, 1.5);
	add_lpf(v, 0, 0, 3, 0.9);
	add_send(v, 0, sDry, 0.7, 0.0);
	add_send(v, 0, sReverb, 0.4, 0.0);
	return v;
}

voice_t *Strings_create(part_t *part)
{
	voice_t	*v = new_voice(part);
	add_oscillator(v, &waveset_pulse[0], 0, 11.95);
	add_oscillator_add(v, &waveset_pulse[0], 0, 12.05);
	add_envelope(v, &envelope_Strings, 1);
	add_velomul(v, 1, 0);
	add_lpf(v, 0, 0, 1, 0.2);
	add_lpf(v, 0, 0, 1, 0.2);
	add_lpf(v, 0, 2, 1, 0.3);
	add_adder(v, 2, 0, -1.0);
	add_send(v, 0, sDry, 0.1, -0.3);
	add_send(v, 0, sReverb, 0.1, 0.3);

	add_oscillator(v, &waveset_pulse[0], 0, 12.0);
	add_oscillator_add(v, &waveset_pulse[0], 0, 11.9);
	add_envelope(v, &envelope_Strings, 1);
	add_velomul(v, 1, 0);
	add_lpf(v, 0, 0, 1, 0.2);
	add_lpf(v, 0, 0, 1, 0.2);
	add_lpf(v, 0, 2, 1, 0.3);
	add_adder(v, 2, 0, -1.0);
	add_send(v, 0, sDry, 0.1, 0.3);
	add_send(v, 0, sReverb, 0.1, -0.3);
	return v;
}

/*
voice_t *Harmonica_create(part_t *part)
{
	voice_t	*v = new_voice(part);
	add_oscillator(v, &waveset_pulse[1], 0, 0.0);
	add_oscillator_add(v, &waveset_saw, 0, 11.95);
	add_envelope(v, &envelope_Strings, 1);
	add_velomul(v, 1, 0);
	add_lpf(v, 0, 0, 1, 0.2);
	add_send(v, 0, sDry, 0.3, 0.0);
	add_send(v, 0, sReverb, 0.5, 0.0);
	return v;
}
*/

void init_instruments()
{
	int i;

	init_oscillators();
	init_envelopes();

	/* The parts */
	for(i = 0; i < 6; ++i)
		p_Strings[i] = add_part(Strings_create, vaMono);
	p_HiHat = add_part(HiHat_create, vaMono);
	p_Bass = add_part(Bass_create, vaMono);
	p_LoopSynth = add_part(LoopSynth_create, vaMono);
	p_BassDrum = add_part(BassDrum_create, vaMono);
//	p_Strings = add_part(Strings_create, vaPoly);
}

void cleanup_instruments()
{
	cleanup_envelopes();
	cleanup_oscillators();
}



/*====================================================================
 --------------------------------------------------------------------
	THE SONG
 --------------------------------------------------------------------

	This is where the notes and other events of the song
	are generated; ie where the score to play the
	instruments is created.
	
====================================================================*/

/* Samples per pulse (PPQN) */
#define	PULSE		160	//184

/* Note lengths; 1/N */
#define	N0		0
#define	P1		(PULSE)
#define	P5		(PULSE * 5)
#define	P8		(PULSE * 8)
#define	N32		(PULSE * 15)
#define	N16		(PULSE * 30)
#define	N8		(PULSE * 60)
#define	N4		(PULSE * 120)
#define	N2		(PULSE * 240)
#define	N1		(PULSE * 480)

void slide(part_t *p, int t, float start, float end, int duration, int interval)
{
	float val = start;
	float k = (end - start) / duration * interval;
	int endtime = t + duration;
	while(1)
	{
		Bend(p, t, val);
		duration -= interval;
		if(duration<0)
		{
			Bend(p, endtime, end);
			return;
		}
		val += k;
		t += interval;
	}
}

int	t = N4;
int	interval = P1;
int	transpose = 0;
float	velocity = 1.0;

#define	NOTE(p,x,y)	({ NoteOn(p, t, (x) + transpose, velocity); t += y; })
#define	NOTEV(p,x,y,z)	({ NoteOn(p, t, (x) + transpose, z); t += y; })
#define	OFF(p)		NoteOff(p, t, 0.5)
#define PAUSE(x)	t += x

#define	BEND(p,x)	Bend(p, t, x)
#define SLIDE(p,s,e,d)	({ slide(p, t, s, e, d, interval); t += d; })

void beats_1(part_t *p, int fadein)
{
	int	i;
	BEND(p, 0);
	for(i = 0; i<32; ++i)
	{
		if(fadein)
	 		NOTEV(p, 24, N0, i/32.0);
		else
	 		NOTE(p, 24, N0);
		SLIDE(p, 42, 20, P8);
		SLIDE(p, 20, 5, N16-P8);
		SLIDE(p, 5, 0, N8);
 		OFF(p); PAUSE(N16);
	}
}

void hats_1(part_t *p)
{
	int	i;
	BEND(p, 0);
	for(i = 0; i<32; ++i)
	{
		if(i & 1)
			NOTEV(p, 0, P8, 1.0);
		else
			NOTEV(p, 24, P8, 1.0);
		OFF(p); PAUSE(N16-P8);
		if((i & 3) == 0)
		{
			NOTEV(p, 3, 2, 1.5);
			OFF(p);
			PAUSE(P8-2);
		}
		else
			NOTEV(p, 24, P8, 1.0);
		OFF(p); PAUSE(N16-P8);
		NOTEV(p, 24, N16, 2.0);

		if(!(i & 4) && (i & 3) != 2)
		{
			NOTEV(p, 3, 2, 1.5);
			OFF(p);
			PAUSE(N16-2);
		}
		else
		{
			NOTEV(p, 24, N16-P8, 1.0);
			OFF(p); PAUSE(P8);
		}
	}
}

void build_song()
{
	int	i;
	part_t	*p;

#if 1
	/* BassDrum part */
	p = p_BassDrum;
	t = N16;
	velocity = 1.0;
	transpose = 0;
	interval = P1;
		beats_1(p, 1);
		beats_1(p, 0);
	PAUSE(N1);
	End(p, t);

	/* HiHat + Edge part */
	p = p_HiHat;
	t = N16;
	velocity = 1.0;
	transpose = 6;
	interval = P1;
		hats_1(p);
		hats_1(p);
	PAUSE(N1);
	End(p, t);
#endif

	/* Bass part */
	p = p_Bass;
	t = N16;
	velocity = 1.0;
	transpose = 12;
	interval = P1;
	BEND(p, 0);
	for(i = 0; i < 128; ++i)
		NOTE(p, 12 + 12*((i&15) % 3 & 1), N16);
	for(i = 0; i < 128; ++i)
 		NOTE(p, 12 + 12*((i&15) % 3), N16);
	PAUSE(N1);
	End(p, t);
#if 1
	/* Synth loop part */
	p = p_LoopSynth;
	t = N16;
	velocity = 1.0;
	for(i = 0; i<8; ++i)
	{
 		BEND(p, 0);
		transpose = 24 + (i & 1)*12;
		interval = P1;
 		NOTEV(p, 24, N0, 0.6); SLIDE(p, 36, 0, N16);
		NOTE(p, 0, N16);

		NOTEV(p, 60, P5, 0.2);
		NOTE(p, 12, N16-P5);
		NOTE(p, 0, N16);

		NOTEV(p, 72, P5, 0.2);
		NOTE(p, 12, N16-P5);
		NOTE(p, 24, N16);

		NOTE(p, 0, N16);
		NOTE(p, 12, N16);

		if((i & 3) == 3)
			++transpose;

		NOTEV(p, 72, P5, 0.2);
		NOTE(p, 0, N16-P5);
		NOTEV(p, 60, P5, 0.2);
		NOTE(p, 12, N16-P5);

		NOTE(p, 24, N16);
		NOTEV(p, 60, P5, 0.2);
		NOTE(p, 12, N16-P5);

		NOTEV(p, 48, P8, 0.4);
		NOTE(p, 0, N16-P8);
		NOTE(p, 12, N16);

		NOTE(p, 24, N16);
		NOTE(p, 12, N0); SLIDE(p, 0, 12, N16);
		BEND(p, 0);
	}
	PAUSE(N1);
	End(p, t);

	/* Strings part */
	for(i = 0; i < 3; ++i)
	{
		static int chord[3] = { 12, 15, 19 };
		static float bend[3] = { 0, 2, 1 };
		t = N16;
		velocity = 0.5;
		transpose = 36;
		interval = P1;
		BEND(p_Strings[i], 0);

		NOTE(p_Strings[i], chord[i], N1 * 4 - N16);
		SLIDE(p_Strings[i], 0, bend[i], N8);
		PAUSE(N1 * 4 - N16);
		OFF(p_Strings[i]);

		PAUSE(N1);
		End(p_Strings[i], t);
	}
#endif
}



/*====================================================================
 --------------------------------------------------------------------
	MAIN
 --------------------------------------------------------------------

	Just initialize the engine, get the instruments and
	song created, and then generate the output audio.

====================================================================*/

int main(int argc, char *argv[])
{
	short *buffer;

	if( (2==argc) && !strcmp("-h",argv[1]) )
	{
		printf(" --------------------------------------- \n");
		printf("|   David Olofson's Comprez2000 Entry   |\n");
		printf("|   Copyright (C) 2000, David Olofson   |\n");
		printf("  -------------------------------------- \n");
		printf("|   Generates raw 16 bit stereo audio   |\n");
		printf("|         at 44100 Hz to stdout.        |\n");
		printf(" --------------------------------------- \n\n");
		return 0;
	}

	buffer = calloc(BUFFER_FRAMES, 2 * sizeof(short));
	if(!buffer)
	{
		fprintf(stderr,"Failed to allocate 16 bit output buffer!\n");
		exit(1);
	}

	init_engine(44100);

	init_instruments();
	build_song();

	while(!engine_finished())
	{
		engine_run(buffer);
		fwrite(buffer, BUFFER_FRAMES, 2 * sizeof(short), stdout);
	}

	cleanup_instruments();
	cleanup_engine();

	return 0;
}

/*= EOF ============================================================*/
